//    |  /           |
//    ' /   __| _` | __|  _ \   __|
//    . \  |   (   | |   (   |\__ `
//   _|\_\_|  \__,_|\__|\___/ ____/
//                   Multi-Physics
//
//  License:		 BSD License
//					 Kratos default license: kratos/license.txt
//
//  Main authors:    Philipp Bucher (https://github.com/philbucher)
//

// System includes
#include <algorithm>
#include <thread>
#include <chrono>

// External includes

// Project includes
#include "includes/kratos_filesystem.h"

namespace Kratos {
namespace filesystem {

bool exists(const std::string& rPath)
{
    return std::filesystem::exists(rPath);
}


bool is_regular_file(const std::string& rPath)
{
    return std::filesystem::is_regular_file(rPath);
}


bool is_directory(const std::string& rPath)
{
    return std::filesystem::is_directory(rPath);
}


bool create_directory(const std::string& rPath)
{
    return std::filesystem::create_directory(rPath);
}


bool create_directories(const std::string& rPath)
{
    return std::filesystem::create_directories(rPath);
}


bool remove(const std::string& rPath)
{
    return std::filesystem::remove(rPath);
}


std::uintmax_t remove_all(const std::string& rPath)
{
    return std::filesystem::remove_all(rPath);
}


void rename(const std::string& rPathFrom, const std::string& rPathTo)
{
    return std::filesystem::rename(rPathFrom, rPathTo);
}

std::string parent_path(const std::string& rPath)
{
    return std::filesystem::path(rPath).parent_path().string();
}

std::string filename(const std::string& rPath)
{
    return std::filesystem::path(rPath).filename().string();
}

} // namespace filesystem


namespace FilesystemExtensions {

std::string CurrentWorkingDirectory()
{
    return std::filesystem::current_path().string();
}

std::string JoinPaths(const std::vector<std::string>& rPaths)
{
    auto paths(rPaths); // create local copy

    // first remove empty paths
    paths.erase(std::remove_if(paths.begin(), paths.end(),
                         [](const std::string& s)
                         { return s.empty(); }), paths.end());

    const std::size_t num_paths = paths.size();

    if (num_paths == 0) { return ""; }

    std::string full_path = paths[0];
    if (num_paths > 1) {
        for(std::size_t i=1; i<num_paths; ++i) {
            full_path += "/" + paths[i]; // using portable separator "/"
        }
    }

    return full_path;
}

std::vector<std::string> ListDirectory(const std::string& rPath)
{
    std::vector<std::string> result;
    for (const auto& current_directory : std::filesystem::directory_iterator(rPath)) {
        result.push_back(current_directory.path().string());
    }
    return result;
}

void MPISafeCreateDirectories(const std::string& rPath)
{
    if (!std::filesystem::exists(rPath)) {
        std::filesystem::create_directories(rPath);
    }
    if (!std::filesystem::exists(rPath)) { // wait for the path to appear in the filesystem
        std::this_thread::sleep_for(std::chrono::milliseconds(50));
    }
}

} // namespace FilesystemExtensions
} // namespace Kratos
